@@ -69,7 +69,8 @@ class Server(IServer):
         self._wait_for:     Optional[Tuple[Awaitable, WaitFor]] = None
         self._wait_for_fut: Optional[Future[WaitFor]] = None
 
-        self._pending_who: Deque[str] = deque()
+        self._pending_who:  Deque[str] = deque()
+        self._initial_nick: Optional[str] = None
 
     def hostmask(self) -> str:
         hostmask = self.nickname
@@ -139,6 +140,8 @@ class Server(IServer):
         username = self.params.username or nickname
         realname = self.params.realname or nickname
 
+        self._initial_nick = nickname
+
         # these must remain non-awaited; reading hasn't started yet
         if not self.params.password is None:
             self.send(build("PASS", [self.params.password]))
@@ -165,11 +168,36 @@ class Server(IServer):
         if line.command == "PING":
             await self.send(build("PONG", line.params))
 
+        elif line.command == RPL_ENDOFWHO:
+            chan = self.casefold(line.params[1])
+            if (self._pending_who and
+                    self._pending_who[0] == chan):
+                self._pending_who.pop()
+                await self._next_who()
+
+        elif (line.command == ERR_NICKNAMEINUSE and
+                self._initial_nick is not None):
+            nick = self._initial_nick
+
+            alt_nicks = self.params.alt_nicknames
+            if not alt_nicks:
+                alt_nicks = [nick+"_"*i for i in range(3)]
+
+            for alt_nick in alt_nicks:
+                if await self.send_nick(alt_nick):
+                    break
+            else:
+                self._initial_nick = None
+                await self.send(build("QUIT"))
+
         elif emit is not None:
             if emit.command == "001":
                 await self.send(build("WHO", [self.nickname]))
                 self.set_throttle(THROTTLE_RATE, THROTTLE_TIME)
 
+                if self.params.autojoin:
+                    await self._batch_joins(self.params.autojoin)
+
             elif emit.command == "CAP":
                 if emit.subcommand    == "NEW":
                     await self._cap_ls(emit)
@@ -182,25 +210,35 @@ class Server(IServer):
 
             elif emit.command == "JOIN":
                 if emit.self and not emit.channel is None:
-                    await self.send(build("MODE", [emit.channel.name]))
+                    chan  = emit.channel.name_lower
+                    await self.send(build("MODE", [chan]))
 
-                    self._pending_who.append(emit.channel.name)
+                    modes = "".join(self.isupport.chanmodes.a_modes)
+                    await self.send(build("MODE", [chan, f"+{modes}"]))
+
+                    self._pending_who.append(chan)
                     if len(self._pending_who) == 1:
-                        await self._serial_who()
+                        await self._next_who()
 
         await self.line_read(line)
 
-    async def _serial_who(self):
-        while self._pending_who:
-            next = self._pending_who.popleft()
+    async def _batch_joins(self,
+            channels: List[str],
+            batch_n:  int=10):
+        #TODO: do as many JOINs in one line as we can fit
+        #TODO: channel keys
+
+        for i in range(0, len(channels), batch_n):
+            batch = channels[i:i+batch_n]
+            await self.send(build("JOIN", [",".join(batch)]))
+
+    async def _next_who(self):
+        if self._pending_who:
+            chan = self._pending_who[0]
             if self.isupport.whox:
-                await self.send(self.prepare_whox(next))
+                await self.send(self.prepare_whox(chan))
             else:
-                await self.send(build("WHO", [next]))
-
-            end = Response(RPL_ENDOFWHO, [ANY, Folded(next)])
-            line = await self.wait_for(end)
-
+                await self.send(build("WHO", [chan]))
 
     async def _read_line(self, timeout: float) -> Optional[Line]:
         while True:
@@ -236,7 +274,7 @@ class Server(IServer):
 
     async def _read_lines(self):
         waiting_lines: List[Tuple[Line, Optional[Emit]]] = []
-        sent_ping = True
+        sent_ping = False
         while True:
             now = monotonic()
             timeouts: List[float] = []
@@ -260,9 +298,11 @@ class Server(IServer):
                     if since >= (PING_TIMEOUT*2):
                         raise ServerDisconnectedException()
                     elif not sent_ping:
+                        sent_ping = True
                         await self.send(build("PING", ["hello"]))
                         continue
             else:
+                sent_ping = False
                 emit = self.parse_tokens(line)
 
                 waiting_lines.append((line, emit))
